from fairfetched.get import Papyrus
from fairfetched.standardization.pipe import with_cleaned_mol_descriptors

p = Papyrus.from_version("05.7")
df = p.full_data.head(100).pipe(with_cleaned_mol_descriptors, parallel=False).collect()
print(df)
