from typing import Callable, TypeAlias, TypeVar

from rdkit.Chem import Mol

StrToMolFn: TypeAlias = Callable[[str], Mol | None]
StrToMolFn: TypeAlias = Callable[[str], Mol | None]
StdFn: TypeAlias = Callable[[Mol], Mol | None]
MolToStrFn: TypeAlias = Callable[[Mol], str | None]
RepFns: TypeAlias = MolToStrFn | dict[str, MolToStrFn]


T_in = TypeVar("T_in")
T_out = TypeVar("T_out")
